package org.processmining.plugins.plg.models;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.processmining.framework.annotations.AuthoredType;
import org.processmining.plugins.plg.config.BasicPluginConfiguration;
import org.processmining.plugins.plg.generator.log.PLGLogCollectionGeneratorConfiguration;

/**
 * This class describes a collection of logs
 *
 * @author Andrea Burattin
 */
@AuthoredType(
	typeName = "PLG Logs collection",
	affiliation = BasicPluginConfiguration.AFFILIATION,
	author = BasicPluginConfiguration.AUTHOR,
	email = BasicPluginConfiguration.EMAIL
)
public class PLGLogCollection {

	private PLGLogCollectionGeneratorConfiguration configuration;
	private ConcurrentHashMap<PlgProcess, Set<PLGLog>> collection;


	/**
	 * Basic constructor
	 *
	 * @param size the expected size of the collection
	 */
	public PLGLogCollection(int size) {
		collection = new ConcurrentHashMap<PlgProcess, Set<PLGLog>>(size);
	}


	/**
	 * Method to set the current configuration that generated this collection
	 *
	 * @param configuration the configuration to set
	 */
	public void setConfiguration(PLGLogCollectionGeneratorConfiguration configuration) {
		this.configuration = configuration;
	}


	/**
	 * Method to get the current configuration that generated this collection
	 *
	 * @return the configuration
	 */
	public PLGLogCollectionGeneratorConfiguration getConfiguration() {
		return configuration;
	}


	/**
	 * Method to add a log to the current collection
	 *
	 * @param p the process that generated the log
	 * @param l the log to add
	 * @return the new set of logs associated to the given process
	 */
	public Set<PLGLog> put(PlgProcess p, PLGLog l) {
		Set<PLGLog> s = collection.get(p);
		if (s == null) {
			s = new HashSet<PLGLog>();
			collection.put(p, s);
		}
		s.add(l);
		return s;
	}


	/**
	 * This method returns the size of the collection
	 *
	 * @return the size of the collection
	 */
	public int size() {
		return collection.size();
	}


	/**
	 * This method to get the set of all the processes with a log
	 *
	 * @return the set of processes
	 */
	public Set<PlgProcess> keySet() {
		return collection.keySet();
	}


	/**
	 * This method to get the set of all the logs for a given process
	 *
	 * @param p the given process
	 * @return the set of logs
	 */
	public Set<PLGLog> get(PlgProcess p) {
		return collection.get(p);
	}
}
